import numpy as np


def nonlinear_transformation(x, transformation='identity'):
    """Transforms marginal of 2D array according to the required function.
    
    Parameters
    ----------
    x : array-like of shape (n_samples, 2)
        Data to transform.
    transformation : string, default='identity'
        One of 'identity', 'square', 'cube' or 'log'. 
        Transform data according to this function. 
        If not specified, or different than the above, return the input array
    
    Returns
    ----------
    x : array-like of shape (n_samples, 2)
        The transformed array.
    """    
    if transformation == 'identity':
        x[:, 1] = x[:, 1]
    elif transformation == 'square':
        x[:, 1] = x[:, 1] + 0.1 * x[:, 1]**2 
    elif transformation == 'cube':
        x[:, 1] = x[:, 1] + 0.5 * x[:, 1]**3 
    elif transformation == 'log':
        x[:, 1] = np.log(x[:, 1]+5.0)      
    return x


import numpy as np
from sklearn.utils import check_random_state
from sklearn.mixture._gaussian_mixture import _estimate_gaussian_covariances_full
from sklearn import cluster
from scipy.special import gamma


class Inits:
    """Generic class to deal with the different initializations. See initialize_parameters 
    for more details, and for all parameters that need to be passed.
    """
    def __init__(self, X, random_state=None, n_components=1, 
                 init_type='random_sklearn', scale=None, reg_covar=0):
        self.X = X
        self.n_samples, self.n_features = self.X.shape
        self.random_state = check_random_state(random_state)
        self.n_components = n_components
        self.init_type = init_type
        if scale is None and (self.init_type=='random' or self.init_type=='minmax'):
            self.scale = self.calculate_scale() 
        elif scale is None and self.init_type=='randomized_kmeans':
            self.scale = 1.0
        else:
            self.scale = scale
        self.reg_covar = reg_covar             

    def calculate_scale(self):
        """Calculate scale prefactor ('s') for the 'random' and 'minmax' initializations.

        Returns
        ----------
        scale : float
            Scale prefactor, to be used to initialize the GMM.
        """
        min_pos = self.X.min(axis=0)
        max_pos = self.X.max(axis=0)
        vol_data = np.prod(max_pos-min_pos)
        # radius of n-ball given its volume; see https://en.wikipedia.org/wiki/Volume_of_an_n-ball
        scale = (vol_data / self.n_components * gamma(self.n_features*0.5 + 1))**(1/self.n_features) / np.sqrt(np.pi)
        return scale
    
    def calculate_parameters_from_responsibilities(self, responsibilities):
        """Calculate GMM parameters given the responsibilities of each sample. 
        The responsibilities are the probabilities that each sample belongs to each component.
        Weights, means and covariances are calculated as in an M-step.

        Parameters
        ----------
        responsibilities : array-like of shape (n_samples, n_components)
            The responsibilities.     

        Returns
        ----------
        weights : array, shape (n_components, 1)
            The initial weights of the GMM model.
        means : array, shape (n_components, n_features)
            The initial means of the GMM model.        
        covariances : array, shape (n_components, n_features, n_features)
            The initial covariance matrices of the GMM model.        
        """
        # we add a small epsilon to avoid underflows
        nk = responsibilities.sum(axis=0) + 10 * np.finfo(responsibilities.dtype).eps
        weights = nk/self.n_samples
        means = np.dot(responsibilities.T, self.X) / nk[:, np.newaxis]
        # we don't add any regularization  to the diagonal of the covariance matrices here
        covariances = _estimate_gaussian_covariances_full(responsibilities, self.X, nk, means, reg_covar=0)        
        return weights, means, covariances
    
    def add_regularization(self, covariances):
        """Add regularization to the diagonal of the covariance matrices, to avoid singularities.
        
        Parameters
        ----------
        covariances : array, shape (n_components, n_features, n_features)
            The unregularized covariance matrices of the GMM model.  

        Returns
        ---------- 
        covariances : array, shape (n_components, n_features, n_features)
            The regularized covariance matrices of the GMM model.  
        """
        regularization = np.repeat((self.reg_covar*np.eye(self.n_features))[np.newaxis, :, :], self.n_components, axis=0)
        covariances += regularization   
        return covariances   
    
    
class RandomInit(Inits):
    def calculate_parameters(self):
        weights = np.repeat(1/self.n_components, self.n_components)
        # initialize components around data points with uncertainty s
        refs = self.random_state.randint(0, self.n_samples, size=self.n_components)
        means = self.X[refs] + self.random_state.multivariate_normal(np.zeros(self.n_features), 
                                                                     self.scale**2 * np.eye(self.n_features), 
                                                                     size=self.n_components)
        covariances = np.repeat(self.scale**2 * np.eye(self.n_features)[np.newaxis, :, :], self.n_components, axis=0)
        return weights, means, covariances

    
class KmeansInit(Inits):
    def calculate_parameters(self):
        from scipy.cluster.vq import kmeans2
        _, label = kmeans2(self.X, self.n_components, seed=self.random_state)
        weights = np.zeros(self.n_components)
        means = np.zeros((self.n_components, self.n_features))
        covariances = np.zeros((self.n_components, self.n_features, self.n_features))
        for k in range(self.n_components):
            mask = (label == k)
            weights[k] = mask.sum() / len(self.X)
            means[k,:] = self.X[mask].mean(axis=0)
            d_m = self.X[mask] - means[k,:] 
            # funny way of saying: for each point i, do the outer product of d_m with its transpose and sum over i
            # in other words, simply the definition of sample covariance matrix
            covariances[k,:,:] = (d_m[:, :, None] * d_m[:, None, :]).sum(axis=0) / len(self.X)
        return weights, means, covariances
    
    
class RandomizedKmeansInit(Inits):
    def calculate_parameters(self):
        from scipy.cluster.vq import kmeans2
        _, label = kmeans2(self.X, self.n_components, seed=self.random_state)
        weights = np.zeros(self.n_components)
        means = np.zeros((self.n_components, self.n_features))
        covariances = np.zeros((self.n_components, self.n_features, self.n_features))
        for k in range(self.n_components):
            mask = (label == k)
            weights[k] = mask.sum() / len(self.X)
            means[k,:] = self.X[mask].mean(axis=0)
            d_m = self.X[mask] - means[k,:] 
            # funny way of saying: for each point i, do the outer product of d_m with its transpose and sum over i
            # in other words, simply the definition of sample covariance matrix
            covariances[k,:,:] = (d_m[:, :, None] * d_m[:, None, :]).sum(axis=0) / len(self.X)           
        # after running k-means, we push the means according to their covariance;
        # this should reduce the risks of getting stuck in local optima
        for k in range(self.n_components):
            means[k,:] = means[k,:] + self.scale*self.random_state.multivariate_normal(np.zeros(self.n_features), 
                                                                     covariances[k,:,:], size=1)
        return weights, means, covariances    
    

class MinMaxInit(Inits):
    def calculate_parameters(self):
        weights = np.repeat(1/self.n_components, self.n_components)
        min_pos = self.X.min(axis=0)
        max_pos = self.X.max(axis=0)
        means = min_pos + (max_pos-min_pos)*self.random_state.rand(self.n_components, self.n_features)
        covariances = np.repeat(self.scale**2 * np.eye(self.n_features)[np.newaxis, :, :], self.n_components, axis=0)
        return weights, means, covariances

    
class RandomSklearnInit(Inits):
    def calculate_parameters(self):
        responsibilities = self.random_state.rand(self.n_samples, self.n_components)
        responsibilities /= responsibilities.sum(axis=1)[:, np.newaxis]
        weights, means, covariances = self.calculate_parameters_from_responsibilities(responsibilities)    
        return weights, means, covariances
    
    
class KmeansSklearnInit(Inits):
    def calculate_parameters(self):
        responsibilities = np.zeros((self.n_samples, self.n_components))
        label = (
            cluster.KMeans(
                n_clusters=self.n_components, n_init=1, random_state=self.random_state
            )
            .fit(self.X)
            .labels_
        )
        responsibilities[np.arange(self.n_samples), label] = 1
        weights, means, covariances = self.calculate_parameters_from_responsibilities(responsibilities)    
        return weights, means, covariances
    
    
_init_type_to_class_map = {'random': RandomInit, 'kmeans': KmeansInit, 'randomized_kmeans': RandomizedKmeansInit,
                           'minmax': MinMaxInit, 'random_sklearn': RandomSklearnInit, 'kmeans_sklearn': KmeansSklearnInit}

def initialize_parameters(X, random_state=None, n_components=1, init_type='random_sklearn', scale=None, reg_covar=0):
    """Initialize the Gaussian mixture model (GMM) parameters (weights, means, covariances and precisions).
    
    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)
        The data based on which the GMM initialization is calculated.    
    random_state : int, default=None
        A random seed used for the method chosen to initialize the parameters.
    n_components : int, default=1
        Number of GMM components to initialize.
    init_type : {'random', 'minmax', 'kmeans', 'randomized_kmeans', 'random_sklearn', 'kmeans_sklearn'}, default='random_sklearn'
        The method used to initialize the weights, the means, the covariances and the precisions.
        Must be one of:
            'random': weights are set uniformly, covariances are proportional to identity (with prefactor s^2). 
            For each mean, a data sample is selected at random, and a multivariate Gaussian with variance s^2 offset is added.
            'minmax': same as above, but means are distributed randomly over the range that is covered by data.
            'kmeans': k-means clustering run as in Algorithm 1 from Bloemer & Bujna (arXiv:1312.5946), 
            as implemented by Melchior & Goulding (arXiv:1611.05806).
            'randomized_kmeans': same as k-means, but then the means are pushed around for increased likelihood 
            of avoiding local minima. The 'scale' is used to quantify the mean push, and can be specified by the user (defaults to 1).
            'random_sklearn': responsibilities are initialized randomly, i.e. every point is randomly assigned to a component.
            Weights, means and covariances are then calculated by performing an M-step on these responsibilities:
            this means that weights are calculated as the average probability that a sample belongs to each components, while
            means and covariance matrices are the weighted mean and covariance of the samples.
            'kmeans_sklearn': responsibilities are initialized using k-means, as implemented by sklearn. 
            The rest is as in 'random_sklearn'.
            Note we are currently not including the sklearn initializations available from v1.1.
    scale : float, default=None
        If set, sets component variances in the 'random' and 'minmax' cases. 
        If scale is not given, it will be set such that the volume of all components completely fills the space covered by data.
    reg_covar : float, default=0
        Constant regularisation term added to the diagonal of each covariance matrix, to avoid singular matrices.
        In GMM-MI, the regularisation term is added during training only, and not in the initializations.
        
    Returns
    ----------
    weights : array, shape (n_components, 1)
        The initial weights of the GMM model.
    means : array, shape (n_components, n_features)
        The initial means of the GMM model.        
    covariances : array, shape (n_components, n_features, n_features)
        The initial covariance matrices of the GMM model.  
    precisions : array, shape (n_components, n_features, n_features)
        The initial precision matrices of the GMM model (inverse of covariance).  
    """
    if init_type in ['random', 'kmeans', 'randomized_kmeans', 'minmax', 'random_sklearn', 'kmeans_sklearn']:
        InitClass = _init_type_to_class_map[init_type]
        init_class = InitClass(X=X, random_state=random_state, n_components=n_components, 
                               init_type=init_type, scale=scale, reg_covar=reg_covar)
        weights, means, covariances = init_class.calculate_parameters()
        covariances = init_class.add_regularization(covariances=covariances)
    else:
        raise ValueError(f"Initialization type not known. It should be one of "
                         f"'random', 'kmeans', 'randomized_kmeans', 'minmax', 'random_sklearn', 'kmeans_sklearn'; found '{init_type}'") 

    # all matrices can be inverted at once
    precisions = np.linalg.inv(covariances)
    return weights, means, covariances, precisions



class GMMFitParamHolder:
    """Container class to provide the hyperparameters pertaining to all GMM fits.
    See mi.py for the meaning of each parameter.
    """
    def __init__(self, init_type="random_sklearn", scale=None, 
                 threshold_fit=1e-5, reg_covar=1e-15, max_iter=10000):
        self.init_type = init_type
        self.scale = scale
        self.threshold_fit = threshold_fit
        self.reg_covar = reg_covar
        self.max_iter = max_iter

class SelectComponentsParamHolder:
    """Container class to provide the hyperparameters pertaining to cross-validation
    and selecting the number of GMM components. See mi.py for the meaning of each parameter.
    """
    def __init__(self, n_inits=3, n_folds=2, metric_method='valid',
                 threshold_components=1e-5, patience=1, max_components=50,
                 fixed_components_number=0):
        self.n_inits = n_inits
        self.n_folds = n_folds
        assert metric_method == 'valid' or metric_method == 'aic' or metric_method == 'bic', f"metric_method must be either 'valid', 'aic' or 'bic, found '{metric_method}'"
        self.metric_method = metric_method
        assert threshold_components >= 0, f"`threshold_components` must be a non-negative number, found '{threshold_components}'"
        self.threshold_components = threshold_components
        assert patience >= 1, f"patience should be at least 1, found {patience}."
        self.patience = patience
        self.max_components = max_components
        self.fixed_components_number = fixed_components_number
        self.fixed_components = True if self.fixed_components_number > 0 else False

class MIDistParamHolder:
    """Container class to provide the hyperparameters pertaining to the MI distribution.
    See mi.py for the meaning of each parameter.
    """
    def __init__(self, n_bootstrap=50, MI_method='MC', 
                MC_samples=1e5):
        self.n_bootstrap = n_bootstrap
        self.MI_method = MI_method
        self.MC_samples = MC_samples

        


import numpy as np
from scipy.special import psi 


def calculate_MI_D1_analytical(covariance_matrix):
    """Calculate the mutual information (MI) for a single bivariate Gaussian using the analytical formula.
    In this simple setting, MI only depends on the correlation coefficient between the two variables.
    
    Parameters
    ----------
    covariance_matrix : array-like of shape (2, 2)
        Covariance matrix of the bivariate Gaussian.
    
    Returns
    ----------
    MI : float
        The value of mutual information in nat.
    """
    assert covariance_matrix.shape==(2, 2), "Covariance matrix must be of shape (2, 2)."
    MI = -0.5*np.log(1-(covariance_matrix[0, 1] / (np.sqrt(covariance_matrix[0, 0]*covariance_matrix[1, 1])))**2)
    return MI


def calculate_MI_gammaexp_analytical(alpha):
    """Calculate the mutual information (MI) for a gamma-exponential distribution, with free parameter \alpha.
    
    Parameters
    ----------
    alpha : float
        Free parameter of the distribution.
    
    Returns
    ----------
    MI : float
        The value of mutual information in nat.
    """
    # psi is the digamma function
    return psi(alpha+1) - np.log(alpha)


def calculate_MI_weinman_analytical(alpha):
    """Calculate the mutual information (MI) for an ordered Weinman exponential distribution, with free parameter \alpha.
    
    Parameters
    ----------
    alpha : float
        Free parameter of the distribution.
    
    Returns
    ----------
    MI : float
        The value of mutual information in nat.
    """
    # psi is the digamma function
    if 0 < alpha < 0.5:
        return -np.log(2*alpha / (1-2*alpha)) + psi( 1 / (1-2*alpha)) -psi(1)
    elif alpha == 0.5:
        return -psi(1)
    else:
        return np.log( (2*alpha-1)  / (2*alpha)) + psi( (2*alpha) / (2*alpha-1)) -psi(1)



import numpy as np
from scipy.stats import multivariate_normal
from scipy.special import logsumexp
from scipy import linalg
import scipy.integrate as integrate
from sklearn.mixture import GaussianMixture as GMM
from sklearn.mixture._gaussian_mixture import _estimate_log_gaussian_prob, _compute_precision_cholesky, _estimate_gaussian_covariances_full
from sklearn.utils import check_random_state
from sklearn.exceptions import ConvergenceWarning
import warnings


class GMMWithMI(GMM):
    """
    Custom Gaussian mixture model (GMM) class built on the sklearn GaussianMixture class. 
    Its main extra feature is the mutual information (MI) estimation, either with MC integration
    or a quadrature method (in 2D). It also has three extra features:
        - allows to work with a GMM with fixed parameters, without fitting them first.
        - the fit_predict function can take as input a validation set, so that the 
        validation log-likelihood can also be tracked.
        - initializations are dealt with separately, so init_params is not used explicitly.
        We recommend calculating the initial parameters with the provided utilities in
        initializations.py, and then providing them as input to this class.
    Most methods and attributes are the same as GaussianMixture; only differences are highlighted here.
    
    Parameters
    ----------
    threshold_fit : float, default=1e-5
        The log-likelihood threshold on each GMM fit used to choose when to stop training.
        Smaller values will improve the fit quality and reduce the chances of stopping at a local optimum,
        while making the code considerably slower. This is equivalent to `tol` in sklearn GMMs; only
        changed to improve clarity.
    weights_init : array-like of shape (n_components, ), default=np.array([1.])
        The user-provided initial weights.
        If None, a single weight is initialized as 1.
    means_init : array-like of shape (n_components, n_features), default=np.array([0.])
        The user-provided initial means,
        If None, a single mean is initialized at 0.
    precisions_init : array-like, default=np.array([1.]).reshape(-1, 1, 1)
        The user-provided initial precisions (inverse of the covariance
        matrices).
        If None, a single precision is initialized as 1.
    covariances_init : array-like, default=None
        The user-provided initial covariances (inverse of the precision matrices).
        If None, covariances are initialized as inverse of initial precision matrices.
    
    Attributes
    ----------
    weights_ : array-like of shape (n_components,)
        The weights of each mixture components. Initially set to the initial values.
    means_ : array-like of shape (n_components, n_features)
        The mean of each mixture component. Initially set to the initial values.
    covariances_ : array-like
        The covariance of each mixture component. Initially set to the initial values.
        The shape depends on `covariance_type`::
            (n_components,)                        if 'spherical',
            (n_features, n_features)               if 'tied',
            (n_components, n_features)             if 'diag',
            (n_components, n_features, n_features) if 'full'
    precisions_ : array-like
        The precision matrices for each component in the mixture. Initially set to the initial values.
        A precision matrix is the inverse of a covariance matrix. A covariance matrix is
        symmetric positive definite so the mixture of Gaussian can be
        equivalently parameterized by the precision matrices. Storing the
        precision matrices instead of the covariance matrices makes it more
        efficient to compute the log-likelihood of new samples at test time.
        The shape depends on `covariance_type`::
            (n_components,)                        if 'spherical',
            (n_features, n_features)               if 'tied',
            (n_components, n_features)             if 'diag',
            (n_components, n_features, n_features) if 'full'
    precisions_cholesky_ : array-like
        The Cholesky decomposition of the precision matrices of each mixture component. 
    train_loss : list
        Contains the log-likelihood on training data as a function of iteration.
        Only created and filled if a validation set is provided when fitting the model.
    val_loss : list
        Contains the log-likelihood on validation data as a function of iteration.
        Only created and filled if a validation set is provided when fitting the model.
    """
    def __init__(self,
                 n_components=1,
                 threshold_fit=1e-5, # this has replaced `tol`
                 reg_covar=1e-15,
                 max_iter=10000,
                 random_state=None,
                 covariance_type='full',
                 weights_init=np.array([1.]),
                 means_init=np.array([0.]),
                 precisions_init=np.array([1.]).reshape(-1, 1, 1),
                 covariances_init=None
                 ):
        super(GMMWithMI, self).__init__(n_components=n_components,
                 covariance_type=covariance_type,
                 reg_covar=reg_covar,
                 max_iter=max_iter,
                 random_state=random_state,
                 weights_init=weights_init,
                 means_init=means_init,
                 precisions_init=precisions_init,
                )
        # having self.means_ (as well as weights and precisions) allows
        # to bypass the check_is_fitted function
        self.means_ = means_init
        if covariances_init is None:
            covariances_init = np.linalg.inv(self.precisions_init)
        self.covariances_ = covariances_init
        self.covariances_init = covariances_init
        self.weights_ = weights_init
        self.covariance_type = covariance_type
        self.precisions_cholesky_ = _compute_precision_cholesky(
                self.covariances_, self.covariance_type
           )
        self.threshold_fit = threshold_fit
    
    def _initialize_parameters(self):
        """Initialize the model parameters. Since we deal with initialization 
        before running the GMM, this simply returns the initial parameters.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None         
        """
        self.weights_ = self.weights_init
        self.means_ = self.means_init       
        self.precisions_cholesky_ = np.array(
            [
                linalg.cholesky(prec_init, lower=True)
                for prec_init in self.precisions_init
            ]
        )        

    def fit(self, X, y=None, val_set=None):
        """Estimate model parameters with the EM algorithm.
        Unlike the sklearn class, we fit only once and deal with the multiple initialisations elsewhere.
        We add the possibility of recording the log-likelihood on the validation set, if provided.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point. Training set.
        y : Ignored
            Not used, present for API consistency by convention.
        val_set : array-like of shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point. Validation set.
            
        Returns
        -------
        self : object
            The fitted mixture.
        """
        self.fit_predict(X, y, val_set)
        return self
    
    def fit_predict(self, X, y=None, val_set=None):
        """Estimate model parameters using X and predict the labels for X.
        The method fits the model once times and sets the parameters with
        which the model has the largest likelihood. Within each
        trial, the method iterates between E-step and M-step for `max_iter`
        times until the change of likelihood or lower bound is less than
        `threshold_fit`, otherwise, a :class:`~sklearn.exceptions.ConvergenceWarning` is
        raised.
        Unlike the sklearn class, we fit only once and deal with the multiple initialisations elsewhere.
        We add the possibility of recording the log-likelihood on the validation set, if provided.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point. Training set.
        y : Ignored
            Not used, present for API consistency by convention.
        val_set : array-like of shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point. Validation set.
            
        Returns
        -------
        labels : array, shape (n_samples,)
            Component labels.
        """
        X = self._validate_data(X, dtype=[np.float64, np.float32], ensure_min_samples=2)
        if X.shape[0] < self.n_components:
            raise ValueError(
                "Expected n_samples >= n_components "
                f"but got n_components = {self.n_components}, "
                f"n_samples = {X.shape[0]}"
            )
        self._check_initial_parameters(X) # these are hyperparameters, not GMM parameters

        do_init = True
        max_lower_bound = -np.inf
        self.converged_ = False
        random_state = check_random_state(self.random_state)
        n_samples, _ = X.shape     

        if do_init:
            self._initialize_parameters()

        lower_bound = -np.inf if do_init else self.lower_bound_
        if val_set is not None:
            self.train_loss, self.val_loss = [], []

        for n_iter in range(1, self.max_iter + 1):
            prev_lower_bound = lower_bound
            log_prob_norm, log_resp = self._e_step(X)
            self._m_step(X, log_resp)
            lower_bound = self._compute_lower_bound(log_resp, log_prob_norm)

            change = lower_bound - prev_lower_bound
            self._print_verbose_msg_iter_end(n_iter, change)

            if val_set is not None:
                log_prob_norm_train, _ = self._e_step(X)
                log_prob_norm_val, _ = self._e_step(val_set)
                self.train_loss.append(log_prob_norm_train)
                self.val_loss.append(log_prob_norm_val)

            if abs(change) < self.threshold_fit:
                self.converged_ = True
                break

        if lower_bound > max_lower_bound or max_lower_bound == -np.inf:
            max_lower_bound = lower_bound
            best_params = self._get_parameters()
            best_n_iter = n_iter

        if not self.converged_:
            warnings.warn(
                "A fit did not converge. "
                "Try different init parameters, "
                "or increase max_iter, decrease threshold_fit "
                "or check for degenerate data.",
                ConvergenceWarning,
            )
            
        self._set_parameters(best_params)
        self.n_iter_ = best_n_iter
        self.lower_bound_ = max_lower_bound


        # Always do a final e-step to guarantee that the labels returned by
        # fit_predict(X) are always consistent with fit(X).predict(X)
        # for any value of max_iter and threshold_fit (and any random_state).
        _, log_resp = self._e_step(X)
        return log_resp.argmax(axis=1)

    def score_samples_marginal(self, X, index):
        """Compute the log-likelihood of each sample for the marginal model, 
        indexed by either 0 (x) or 1 (y).
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            List of n_features-dimensional data points. Each row
            corresponds to a single data point.
        index: integer
            Either 0 (marginal x) or 1 (marginal y).
        
        Returns
        -------
        log_prob : array, shape (n_samples,)
            Log-likelihood of each sample in X under the marginal model.
        """
        # in 1-D the Cholesky decomposition is simply the inverse sqrt of the variance
        oned_cholesky = np.sqrt(1/self.covariances_[:, index, index]).reshape(-1, 1, 1)
        marginal_logprob = _estimate_log_gaussian_prob(
            X, self.means_[:, index].reshape(-1, 1), oned_cholesky, self.covariance_type
        )
        return logsumexp(np.log(self.weights_) + marginal_logprob, axis=1)

    def estimate_MI_MC(self, MC_samples=1e5):
        """Compute the mutual information (MI) associated with a particular GMM model, 
        using MC integration.
        
        Parameters
        ----------
        MC_samples : integer, default=1e5
            Number of Monte Carlo (MC) samples to perform numerical integration of the MI integral.
        
        Returns
        -------
        MI : float
            The value of mutual information.
        """
        points, _ = self.sample(MC_samples)       
        # evaluate the log-likelihood for the joint probability
        joint = self.score_samples(points)
        # and the marginals; index=0 corresponds to x, index=y corresponds to y
        marginal_x = self.score_samples_marginal(points[:, :1], index=0)
        marginal_y = self.score_samples_marginal(points[:, 1:], index=1)
        MI = np.mean(joint - marginal_x - marginal_y)
        self.MI = MI
        return MI

    def estimate_MI_quad(self, tol_int=1.49e-8, limit=np.inf):
        """Compute the mutual information (MI) associated with a particular GMM model, 
        using quadrature integration.
        
        Parameters
        ----------
        tol_int : float, default=1.49e-8
            Integral tolerance; the default value is the one form scipy. 
        limit : float, default=np.inf
            The extrema of the integral to calculate. Usually the whole plane, so defaults to inf.
            Integral goes from -limit to +limit.
        
        Returns
        -------
        MI : float
            The value of mutual information.
        """
        # we create a GMM object to pass to the integral functions
        gmm = GMMWithMI(n_components=self.n_components, weights_init=self.weights_, 
                     means_init=self.means_, covariances_init=self.covariances_)
        entropy_2d = integrate.dblquad(entropy_2d_integrand, -limit, limit, 
                                      lambda x: -limit, lambda x: limit, 
                                      args=[gmm], epsabs=tol_int, epsrel=tol_int)[0]
        entropy_1d_x = integrate.quad(entropy_1d_integrand, -limit, limit, 
                                      args=(gmm, 0), epsabs=tol_int, epsrel=tol_int)[0]
        entropy_1d_y = integrate.quad(entropy_1d_integrand, -limit, limit, 
                                      args=(gmm, 1), epsabs=tol_int, epsrel=tol_int)[0]
        MI = entropy_2d - entropy_1d_x - entropy_1d_y
        self.MI = MI
        return MI
    
    
def log_pdf(y, x, model):
    """Log-likelihood in 2D for a given model
    (typically a Gaussian mixture model, GMM).
        
    Parameters
    ----------
    y : float
        The y variable.
    x : float, 
        The x variable.
    model : instance of class with score_samples method
        The model whose log-likelihood we calculate; typically a GMM.
    
    Returns
    -------
    integrand : float
        The value of the integrand.
    """
    y = np.array(y)
    x = np.array(x)
    X = np.concatenate((y.reshape(1, 1), x.reshape(1, 1))).T
    return model.score_samples(X)


def entropy_2d_integrand(y, x, model):
    """Integrand function of 2D entropy for a given model
    (typically a Gaussian mixture model, GMM).
        
    Parameters
    ----------
    y : float
        The y variable.
    x : float, 
        The x variable.
    model : instance of class with score_samples method
        The model whose entropy we calculate; typically a GMM.
    
    Returns
    -------
    integrand : float
        The value of the integrand.
    """
    logp = log_pdf(y, x, model)
    p = np.exp(logp)
    integrand = p*logp
    return integrand


def entropy_1d_integrand(x, model, index):
    """Integrand function of 1D entropy for a given model
    (typically a Gaussian mixture model, GMM).
        
    Parameters
    ----------
    x : float, 
        The x variable.
    model : instance of class with score_samples method
        The model whose entropy we calculate; typically a GMM.
        Must have a score_samples_marginal method.
    index : int
        The index used to calculate entropy, either 0 (x) or 1 (y).
    
    Returns
    -------
    integrand : float
        The value of the integrand.
    """
    assert index == 0 or index == 1, f"Index must be either 0 (x) or 1 (y); found '{index}'"
    x = np.array(x).reshape(1, 1)
    logp = model.score_samples_marginal(x, index)
    p = np.exp(logp)
    integrand = p*logp
    return integrand
    


import numpy as np
from sklearn.model_selection import KFold
import warnings
from gmm_mi.single_fit import single_fit
from gmm_mi.initializations import initialize_parameters

class CrossValidation:
    """Perform cross-validation (CV) to select the best GMM initialization parameters, 
    and thus avoid local minima in the density estimation.

    Parameters
    ----------
    n_components : int
        Number of GMM components currently being fitted.
    n_folds : int, default=2
        Number of folds.
    n_inits : int, default=3
        Number of initializations.
    init_type : {'random', 'minmax', 'kmeans', 'randomized_kmeans', 'random_sklearn', 'kmeans_sklearn'}, default='random_sklearn'
        The method used to initialize the weights, the means, the covariances and the precisions in each fit.
        See utils.initializations for more details.
    cale : float, default=None
        The scale to use to initialize the GMM parameters. Only used if 'init_type' is 'random', 
        'minmax' or 'randomized_sklearn'.
    reg_covar : float, default=1e-15
        The constant term added to the diagonal of the covariance matrices to avoid singularities.
    threshold_fit : float, default=1e-5
        The log-likelihood threshold on each GMM fit used to choose when to stop training.
        Smaller values will improve the fit quality and reduce the chances of stopping at a local optimum,
        while making the code considerably slower. This is equivalent to `tol` in sklearn GMMs.       
    max_iter : int, default=10000
        The maximum number of iterations in each GMM fit. We aim to stop only based on `threshold_fit`, 
        so it is set to a high value. A warning is raised if this threshold is reached; in that case, 
        simply increase this value, or check that you really need such a small `threshold_fit` value.
    """
    def __init__(self, n_components, n_folds=2, n_inits=3, init_type='random_sklearn', 
                 scale=None, reg_covar=1e-15, threshold_fit=1e-5, max_iter=10000):
        self.n_components = n_components
        self.n_folds = n_folds
        self.n_inits = n_inits
        self.init_type = init_type
        self.scale = scale
        self.reg_covar = reg_covar
        self.threshold_fit = threshold_fit
        self.max_iter = max_iter
        
    def _create_containers(self):
        """Create all arrays and lists to store CV results"""
        # create empty arrays to store validation log-likelihoods
        self.val_scores = np.zeros((self.n_inits, self.n_folds))    
        # create empty arrays for final GMM parameters, and loss curves
        self.all_ws = np.zeros((self.n_inits, self.n_folds, self.n_components))
        self.all_ms = np.zeros((self.n_inits, self.n_folds, self.n_components, self.features))
        self.all_ps = np.zeros((self.n_inits, self.n_folds, self.n_components, self.features, self.features))    
        self.all_lcurves = []
        self.convergence_flags = []
        # random split seed is fixed here, but results should be independent of the exact split
        self.kf = KFold(n_splits=self.n_folds, shuffle=True, random_state=42)
        
    def _run_cross_validation(self, X):        
        """
        Actually run the cross-validation (CV) procedure, filling all arrays and lists with results.
        
        Parameters
        ----------
            X : array-like of shape (n_samples, n_features)
                Data on which CV is performed.
        """
        for random_state in range(self.n_inits):
            # initialize with different seed random_state
            w_init, m_init, c_init, p_init = initialize_parameters(X, random_state=random_state, 
                                                                   n_components=self.n_components, 
                                                                   init_type=self.init_type,
                                                                   scale=self.scale)       
            # perform k-fold CV
            for k_id, (train_indices, valid_indices) in enumerate(self.kf.split(X)):
                X_training = X[train_indices]
                X_validation = X[valid_indices]            
                gmm = single_fit(X=X_training, n_components=self.n_components, reg_covar=self.reg_covar, 
                                 threshold_fit=self.threshold_fit, max_iter=self.max_iter, random_state=random_state, 
                                 w_init=w_init, m_init=m_init, p_init=p_init, val_set=X_validation)
                # we take the mean logL per sample, since folds might have slightly different sizes
                val_score = gmm.score_samples(X_validation).mean()            
                # save current scores, as well as GMM parameters
                self.val_scores[random_state, k_id] = np.copy(val_score)
                self.all_ws[random_state, k_id] = np.copy(gmm.weights_)
                self.all_ms[random_state, k_id] = np.copy(gmm.means_)
                self.all_ps[random_state, k_id] = np.copy(gmm.precisions_)
                # save the loss functions as well
                self.all_lcurves.append(np.copy(gmm.train_loss))
                self.all_lcurves.append(np.copy(gmm.val_loss))
                # save a flag to warn about non-convergence
                convergence_flag = False if gmm.n_iter_ <= 2 else True
                self.convergence_flags.append(convergence_flag)
                
    def fit(self, X):
        """Perform CV on the given data.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data on which CV is performed.
        
        Returns
        -------
        results_dict : dict
            A dictionary containing all results from the CV procedure:
            'best_seed': the initialization seed that led to the highest mean log-likelihood across folds.
            'best_val_score': the corresponding value of the highest mean log-likelihood.
            'best_fold_in_init':  the index of the best fold within the best initialization.
            'all_lcurves': list of list with the training and validation curves.
            'all_ws': all weights of the fitted GMM models.
            'all_ms': all means of the fitted GMM models.
            'all_ps': all precisions of the fitted GMM models.
        """
        self.features = X.shape[1]
        self._create_containers()
        self._run_cross_validation(X)
        # select seed with highest val score across the different inits
        avg_val_scores = np.mean(self.val_scores, axis=1)
        best_seed = np.argmax(avg_val_scores)
        best_val_score = np.max(avg_val_scores)
        # within the best fold, also select the model with the highest validation logL
        best_fold_in_init = np.argmax(self.val_scores[best_seed])    
        results_dict = {'best_seed': best_seed, 'best_val_score': best_val_score, 
                        'best_fold_in_init': best_fold_in_init, 'all_lcurves': self.all_lcurves, 
                        'all_ws': self.all_ws, 'all_ms': self.all_ms, 'all_ps': self.all_ps,
                        'convergence_flags': self.convergence_flags}
        return results_dict  


import numpy as np
from gmm_mi.gmm import GMMWithMI as GMM


def single_fit(X, n_components, reg_covar=1e-15, threshold_fit=1e-5, max_iter=10000, 
                random_state=None, w_init=None, m_init=None, p_init=None, val_set=None):
    """
    Perform a single fit of a GMM on input data.

    Parameters
    ----------
    X : array-like of shape (n_samples, 2)
        Data to fit.
    n_components : int
        Number of GMM components to fit.
    reg_covar : float, default=1e-15
        The constant term added to the diagonal of the covariance matrices to avoid singularities.
    threshold_fit : float, default=1e-5
        The log-likelihood threshold on each GMM fit used to choose when to stop training.
        Smaller values will improve the fit quality and reduce the chances of stopping at a local optimum,
        while making the code considerably slower. This is equivalent to `tol` in sklearn GMMs.       
    max_iter : int, default=10000
        The maximum number of iterations in each GMM fit. We aim to stop only based on `threshold_fit`, 
        so it is set to a high value. A warning is raised if this threshold is reached; in that case, 
        simply increase this value, or check that you really need such a small `threshold_fit` value.
    random_state : int, default=None
        Random seed used to initialize the GMM model. 
        If initial GMM parameters are provided, used only to fix the trained model samples across trials.
    w_init : array-like of shape (n_components)
        Initial GMM weights.
    m_init : array-like of shape (n_components, 2)
        Initial GMM means.
    p_init : array-like of shape ( n_components, 2, 2)
        Initial GMM precisions.
    val_set : array-like of shape (n_samples, 2), default=None
        Validation set. If provided, also loss curves are considered.

    Returns
    ----------
    gmm : instance of GMM class
        The fitted GMM model.
    """
    gmm = GMM(n_components=n_components, reg_covar=reg_covar, 
                threshold_fit=threshold_fit, max_iter=max_iter, 
                random_state=random_state, weights_init=w_init, 
                means_init=m_init, precisions_init=p_init).fit(X, val_set=val_set)
    return gmm


import numpy as np
import warnings
from sklearn.exceptions import ConvergenceWarning
from tqdm import tqdm
from gmm_mi.cross_validation import CrossValidation
from gmm_mi.single_fit import single_fit
from gmm_mi.param_holders import GMMFitParamHolder, SelectComponentsParamHolder, MIDistParamHolder


class EstimateMI:
    """Class to calculate mutual information (MI) distribution on 2D data, using Gaussian mixture models (GMMs).
    The main method is `fit` for continuous data, and `fit_categorical` for continuous-categorical data.
    The constructor parameters are mostly inherited from three classes: GMMFitParamHolder, and MIDistParamHolder

    Parameters
    ----------
    (inherited from GMMFitParamHolder, passed as gmm_fit_params)
    init_type : {'random', 'minmax', 'kmeans', 'randomized_kmeans', 'random_sklearn', 'kmeans_sklearn'}, default='random_sklearn'
        The method used to initialize the weights, the means, the covariances and the precisions in each CV fit.
        See utils.initializations for more details.
    scale : float, default=None
        The scale used for 'random', 'minmax' and 'randomized_kmeans' initializations. Not used in all other cases.
        See utils.initializations for more details.
    threshold_fit : float, default=1e-5
        The log-likelihood threshold on each GMM fit used to choose when to stop training.
        Smaller values will improve the fit quality and reduce the chances of stopping at a local optimum,
        while making the code considerably slower. This is equivalent to `tol` in sklearn GMMs.      
    reg_covar : float, default=1e-15
        The constant term added to the diagonal of the covariance matrices to avoid singularities.
        Smaller values will increase the chances of singular matrices, but will have a smaller impact
        on the final MI estimates. 
    max_iter : int, default=10000
        The maximum number of iterations in each GMM fit. We aim to stop only based on `threshold_fit`, 
        so it is set to a high value. A warning is raised if this threshold is reached; in that case, 
        simply increase this value, or check that you really need such a small `threshold_fit` value.
        
    (inherited from ChooseComponentParamHolder, passed as select_component_params)    
    n_inits : int, default=3
        Number of initializations used to find the best initialization parameters.
        Higher values will decrease the chances of stopping at a local optimum, while making the code slower.
    n_folds : int, default=2
        Number of folds in the cross-validation (CV) performed to find the best initialization parameters.
        A good value should ensure each fold has enough samples to be representative of your training set.
    metric_method : {'valid', 'aic', 'bic'}, default='valid'
        Metric used to select the optimal number of GMM components to perform density estimation.
        Must be one of:
            'valid': stop adding components when the validation log-likelihood stops increasing.
            'aic': stop adding components when the Akaike information criterion (AIC) stops decreasing.
            'bic': same as 'aic' but with the Bayesian information criterion (BIC).
    threshold_components : float, default=1e-5
        The metric threshold to decide when to stop adding GMM components. In other words, GMM-MI stops 
        adding components either when the metric gets worse, or when the improvement in the metric value
        is less than this threshold.
        Smaller values ensure that enough components are considered and thus that the data distribution is 
        correctly captured, while taking longer to converge and possibly insignificantly changing the 
        final value of MI.
    patience : int, default=1
        Number of extra components to "wait" until convergence is declared. Must be at least 1.
        Same concept as patience when training a neural network. Higher value will fit models
        with higher numbers of GMM components, while taking longer to converge.        
    max_components : int, default=50
        Maximum number of GMM components that is going to be tested. Hopefully stop much earlier than this.
        A warning is raised if this number of components is used; if so, you might want to use a different
        `metric_method`.
        
    (inherited from MIDistParamHolder, passed as mi_dist_params)            
    n_bootstrap : int, default=50 
        Number of bootstrap realisations to consider to obtain the MI uncertainty.
        Higher values will return a better estimate of the MI uncertainty, and
        will make the MI distribution more Gaussian-like, but the code will take longer.
        If < 1, do not perform bootstrap and actually just do a single fit on the entire dataset.
    MI_method : {'MC', 'quad'}, default='MC' 
        Method to calculate the MI integral. Must be one of:
            'MC': use Monte Carlo integration with MC_samples samples.
            'quad': use quadrature integration, as implemented in scipy, with default parameters.
    MC_samples : int, default=1e5
        Number of MC samples to use to estimate the MI integral. Only used if MI_method == 'MC'.
        Higher values will return less noisy estimates of MI, but will take longer.
    fixed_components_number : int, default=0
        The number of GMM components to use. If 0 (default), will do cross-validation to select 
        the number of components, and ignore this. Else, use this specified number of components,
        and ignore cross-validation.
        
    Attributes
    ----------
    converged : bool
        Flag to check whether we found the optimal number of components. Initially False.
    best_metric : float
        Metric value that is tracked to decide convergence with respect to the number of components.
        This can be either the validation log-likelihood, the AIC or BIC.
    patience_counter : int
        Counter to keep track of patience. Initially 0.
    results_dict : dict
        To keep track of the cross-validation results, and decide convergence.
    fixed_components : bool
        Set to True only if the user fixes a number of GMM components > 0.
    """
    def __init__(self, gmm_fit_params=None, select_components_params=None, mi_dist_params=None): 
        self.gmm_fit_params = gmm_fit_params if gmm_fit_params else GMMFitParamHolder()
        self.sel_comp_params = select_components_params if select_components_params else SelectComponentsParamHolder()
        self.mi_dist_params = mi_dist_params if mi_dist_params else MIDistParamHolder()
       
        self.converged = False
        self.best_metric = -np.inf
        self.patience_counter = 0
        self.results_dict = {}    
    
    def __getattr__(self, attr):
        """Access all hyperparameter classes methods and attributes."""
        try:
            return getattr(self.gmm_fit_params, attr)
        except:
            pass
        try:
            return getattr(self.sel_comp_params, attr)
        except:
            pass
        try:
            return getattr(self.mi_dist_params, attr)
        except:
            pass
    
    def _select_best_metric(self, n_components):
        """Select best metric to choose the number of GMM components.

        Parameters
        ----------
        n_components : int
            Number of GMM components currently being fitted. 

        Returns
        ----------
        metric : float
            The value of the metric selected to choose the number of components.
        """    
        if self.metric_method == 'aic' or self.metric_method == 'bic':
            # in this case we need to re-fit the dataset; we start from the final point
            _, _, w_init, m_init, p_init = self._extract_best_parameters(n_components=n_components, 
                                                                   fixed_components=False, 
                                                                   patience=0)
            gmm = single_fit(X=self.X, n_components=n_components, reg_covar=self.reg_covar, 
                             threshold_fit=self.threshold_fit, max_iter=self.max_iter, 
                             w_init=w_init, m_init=m_init, p_init=p_init)  
            # this is an extra fit we make, so we also check if this converged
            convergence_flag = False if gmm.n_iter_ <= 2 else True
            self.results_dict[n_components]['convergence_flags'].append(convergence_flag)            
            if self.metric_method == 'aic':
                # negative since we maximise the metric
                metric = -gmm.aic(self.X)
            elif self.metric_method == 'bic':
                metric = -gmm.bic(self.X)
        elif self.metric_method == 'valid':
            metric = self.results_dict[n_components]['best_val_score']
        else:
            raise ValueError(f"metric_method must be either 'valid', 'aic' or 'bic, found '{self.metric_method}'")
        return metric        

    def _extract_best_parameters(self, n_components, fixed_components, patience):
        """Extract best parameters relative to cross-validation.

        Parameters
        ----------
        n_components : int
            Number of GMM components to fit. 
        fixed_components : bool
            Fix the number of GMM components to use for density estimation.
            In this instance, only used to decide if patience is used or not.
        patience : int
            Number of extra components to "wait" until convergence is declared. 
            Decides how far back to look to select the number of components. Only used if fixed_components is not True.

        Returns
        ----------
        best_components : int
            Best number of components to fit (either the input ones, or fewer if it's during patience).
        best_seed : int
            Best seed of the initialization that led to the highest validation log-likelihood.
        w_init : array-like of shape (n_components)
            Best GMM weights (based on cross-validation), to be used as initial point of further fits.
        m_init : array-like of shape (n_components, 2)
            Best GMM means (based on cross-validation), to be used as initial point of further fits.
        p_init : array-like of shape (n_components, 2, 2)
            Best GMM precisions (based on cross-validation), to be used as initial point of further fits.
        """    
        best_components = n_components
        if not fixed_components:
            best_components -= patience
        self.lcurves = self.results_dict[best_components]['all_lcurves']
        best_seed = self.results_dict[best_components]['best_seed']
        best_fold_in_init = self.results_dict[best_components]['best_fold_in_init']            
        all_ws = self.results_dict[best_components]['all_ws']
        all_ms = self.results_dict[best_components]['all_ms']
        all_ps = self.results_dict[best_components]['all_ps']
        w_init = all_ws[best_seed, best_fold_in_init]
        m_init = all_ms[best_seed, best_fold_in_init]
        p_init = all_ps[best_seed, best_fold_in_init] 
        return best_components, best_seed, w_init, m_init, p_init

    def _check_convergence(self, n_components):
        """Check if convergence with respect to the number of components has been reached.
        Convergence is declared when the metric has not improved, or when the improvement is
        less than the specified `threshold_components`.

        Parameters
        ----------
        n_components : int
            Number of GMM components being fitted. 
        """    
        if self.metric - self.best_metric > self.threshold_components:
            self.best_metric = self.metric
            if self.verbose:
                print(f'Current number of GMM components: {n_components}. Current metric: {self.best_metric:.3f}. Adding one component...')
        else:
            self.patience_counter += 1
            if self.verbose:
                print(f'Metric change is less than threshold; patience counter increased by 1...')
            if self.patience_counter >= self.patience:
                if self.verbose:
                    print(f'Reached patience limit, stop adding components.')
                self.converged = True

    def _calculate_MI(self, gmm, tol_int=1.49e-8, limit=np.inf):
        """Calculate mutual information (MI) integral given a Gaussian mixture model in 2D.
        Use either Monte Carlo (MC) method, or quadrature method.

        Parameters
        ----------
        gmm : instance of GMM class
            The GMM model whose MI we calculate.
        tol_int : float, default=1.49e-8
            Integral tolerance; the default value is the one form scipy. 
            Only used when MI_method='quad'.
        limit : float, default=np.inf
            The extrema of the integral to calculate. Usually the whole plane, so defaults to inf.
            Only used when MI_method='quad'. Integral goes from -limit to +limit.

        Returns
        ----------
        MI : float
            The value of MI.
        """    
        if self.MI_method == 'MC':
            MI = gmm.estimate_MI_MC(MC_samples=self.MC_samples)
        elif MI_method == 'quad':
            MI = gmm.estimate_MI_quad(tol_int=tol_int, limit=limit)
        return MI

    def _perform_bootstrap(self, n_components, random_state, w_init, m_init, p_init):
        """Perform bootstrap on the given data to calculate the distribution of mutual information (MI)
        in the continuous-continuous case. If n_bootstrap < 1, do only a single fit on the entire dataset.

        Parameters
        ----------
        n_components : int
            Number of GMM components to fit.
        random_state : int, default=None
            Random seed used to initialize the GMM model. 
            If initial GMM parameters are provided, used only to fix the trained model samples across trials.
        w_init : array-like of shape (n_components)
            Initial GMM weights.
        m_init : array-like of shape (n_components, 2)
            Initial GMM means.
        p_init : array-like of shape (n_components, 2, 2)
            Initial GMM precisions.

        Returns
        ----------
        MI_mean : float
            Mean of the MI distribution, in nat.
        MI_std : float
            Standard deviation of the MI distribution, in nat.
        """ 
        if self.n_bootstrap < 1:
            do_bootstrap = False
            self.n_bootstrap = 1 # we will use only one dataset, i.e. all data
        else:
            do_bootstrap = True
        MI_estimates = np.zeros(self.n_bootstrap)
        for n_b in tqdm(range(self.n_bootstrap)):
            if do_bootstrap:
                # we use index n_b to change the seed so that results will be fully reproducible
                rng = np.random.default_rng(n_b)
                X_bs = rng.choice(self.X, self.X.shape[0])
            else:
                X_bs = self.X
            gmm = single_fit(X=X_bs, n_components=n_components, reg_covar=self.reg_covar, 
                             threshold_fit=self.threshold_fit, max_iter=self.max_iter, 
                             random_state=random_state, w_init=w_init, m_init=m_init, p_init=p_init)
            current_MI_estimate = self._calculate_MI(gmm)
            MI_estimates[n_b] = current_MI_estimate
        MI_mean = np.mean(MI_estimates)
        MI_std = np.sqrt(np.var(MI_estimates, ddof=1)) if do_bootstrap else None
        return MI_mean, MI_std

    def fit(self, X, return_lcurves=False, verbose=False):
        """Calculate mutual information (MI) distribution (in nat).
        The first part performs density estimation of the data using GMMs and k-fold cross-validation.
        The second part uses the fitted model to calculate MI, using either Monte Carlo or quadrature methods.
        The MI uncertainty is calculated through bootstrap.

        Parameters
        ----------  
        X : array-like of shape (n_samples, 2)
            Samples from the joint distribution of the two variables whose MI is calculated.
        return_lcurves : bool, default=False
            Whether to return the loss curves or not (for debugging purposes).                          
        verbose : bool, default=False
            Whether to print useful procedural statements.
                
        Returns
        ----------
        MI_mean : float
            Mean of the MI distribution, in nat.
        MI_std : float
            Standard deviation of the MI distribution, in nat.
        loss_curves : list of lists
            Loss curves of the models trained during cross-validation; currently used for debugging.
            Only returned if return_lcurves is true.
        """
        assert X.shape[1] == 2, f"The shape of the data must be (n_samples, 2), found {X.shape}"    
        self.X = X
        self.verbose = verbose
        if self.verbose:
            if not self.fixed_components:
                print('Starting cross-validation procedure to select the number of GMM components...')
            else:
                print(f'Using {self.fixed_components_number} GMM components, as specified in input.')
        for n_components in range(1, self.max_components+1):
            if self.fixed_components:
                if n_components < self.fixed_components_number:
                    continue  
                else:
                    self.converged = True
            current_results_dict = CrossValidation(n_components=n_components, n_folds=self.n_folds, 
                                                   max_iter=self.max_iter, init_type=self.init_type, scale=self.scale,
                                                   n_inits=self.n_inits, threshold_fit=self.threshold_fit, 
                                                   reg_covar=self.reg_covar).fit(self.X)
            self.results_dict[n_components] = current_results_dict
            if not self.converged:
                self.metric = self._select_best_metric(n_components=n_components)
                # store metric, just to make it accessible outside of this class, if needed
                self.results_dict[n_components]['metric'] = self.metric
                self._check_convergence(n_components=n_components)

            if self.converged:
                best_components, best_seed, w_init, m_init, p_init = self._extract_best_parameters(n_components=n_components,    
                                                                                                  fixed_components=self.fixed_components,
                                                                                                  patience=self.patience)
                # these are assigned to self only to possibly plot the final model
                # in `plot_fitted_model`.
                self.best_components = best_components
                self.best_seed = best_seed
                self.w_init = w_init
                self.m_init = m_init
                self.p_init = p_init
                
                if self.verbose:
                    print(f'Convergence reached at {best_components} GMM components.') 
                    print(f'Starting MI integral estimation...') 

                # check if fits actually went on for a good amount of iterations
                if self.fixed_components:
                    convergence_flags = self.results_dict[self.best_components]['convergence_flags'] 
                else:
                    convergence_flags = [self.results_dict[n_c]['convergence_flags'] 
                                         for n_c in range(1, self.best_components+1)]
                # checking if all elements are False; in this case, a warning should be raised
                if not np.sum(convergence_flags): 
                    warnings.warn(
                        f"All CV GMM fits converged only after their second iteration for all components; "
                        "this is usually suspicious, and might be a symptom of a bad fit. "
                        "Plot the loss curves as described in the walkthrough, and try reducing threshold_fit, "
                        "or with a different init_type.",
                        ConvergenceWarning,
                    )  
                    
                # get MI distribution
                MI_mean, MI_std = self._perform_bootstrap(n_components=best_components, random_state=best_seed, 
                                                             w_init=w_init, m_init=m_init, p_init=p_init)
                break

            # in the unlikely case that we have not reached enough GMM components,
            # we raise a ConvergenceWarning 
            if n_components == self.max_components:
                self.reached_max_components = True
                warnings.warn(f"Convergence in the number of GMM components was not reached. "\
                              f"Try increasing max_components or threshold_components, "\
                              f"or decreasing the patience.", ConvergenceWarning)
                best_components, best_seed, w_init, m_init, p_init = self._extract_best_parameters(n_components=self.max_components,    
                                                                                                  fixed_components=True,
                                                                                                  patience=0)
                # these are assigned to self only to possibly plot the final model
                # in `plot_fitted_model`.
                self.best_components = best_components
                self.best_seed = best_seed
                self.w_init = w_init
                self.m_init = m_init
                self.p_init = p_init
                
                # get MI distribution               
                MI_mean, MI_std = self._perform_bootstrap(n_components=best_components, random_state=best_seed, 
                                             w_init=w_init, m_init=m_init, p_init=p_init)                
        
        if self.verbose:
            print('MI estimation completed, returning mean and standard deviation.')
            
        if return_lcurves:
            return MI_mean, MI_std, self.lcurves        
        else:
            return MI_mean, MI_std    
 
    def plot_fitted_model(self, ax=None, **kwargs):
        """Fit model to inout data and plot its contours.
        Only works if the model has been fitted successfully first.
        
        Parameters
        ----------
        ax : instance of the axes.Axes class from pyplot, default=None
            The panel where to plot the samples. 
        kwargs : dictionary
            The extra keyword arguments to pass to the plotting function.
        
        Returns
        -------
        fig: instance of the figure.Figure class from pyplot
            The output figure.
        ax : instance of the axes.Axes class from pyplot
            The output panel.
        """
        assert (self.converged == True or self.reached_max_components == True), \
                                       "You can only plot the fitted model after MI has "\
                                       "been estimated; call .fit() on your data first!"
        from gmm_mi.utils.plotting import plot_gmm_contours
        gmm = single_fit(X=self.X, n_components=self.best_components, reg_covar=self.reg_covar, 
                 threshold_fit=self.threshold_fit, random_state=self.best_seed, max_iter=self.max_iter, 
                 w_init=self.w_init, m_init=self.m_init, p_init=self.p_init)
        fig, ax = plot_gmm_contours(gmm, ax=ax, label='Fitted model', **kwargs)
        return fig, ax
               
    def _calculate_MI_categorical(self):
        """Calculate mutual information (MI) integral given a Gaussian mixture model in 2D.
        Use only Monte Carlo (MC) method. 
        The complete formula can be found in Appendix B of Piras et al. (2022).

        Returns
        -------
        MI : float
            The value of MI.
        """    
        MI = 0 
        for category_value in range(self.category_values):
            samples = self.all_gmms[category_value].sample(self.MC_samples)[0]
            log_p = self.all_gmms[category_value].score_samples(samples)
            p_inner = 0
            for inner_category_value in range(self.category_values):
                p_inner += np.exp(self.all_gmms[inner_category_value].score_samples(samples))
            p = np.log(p_inner/self.category_values)
            MI += np.mean(log_p - p)
        MI /= self.category_values
        return MI
     
    def _perform_bootstrap_categorical(self):
        """Perform bootstrap on the given data to calculate the distribution of mutual information (MI),
        in the categorical-continuous case. If n_bootstrap < 1, do only a single fit on the entire dataset.

        Returns
        ----------
        MI_mean : float
            Mean of the MI distribution, in nat.
        MI_std : float
            Standard deviation of the MI distribution, in nat. None if bootstrap==False.
        """  
        if self.n_bootstrap < 1:
            do_bootstrap = False
            self.n_bootstrap = 1 # we will use only one dataset, i.e. all data
        else:
            do_bootstrap = True
        MI_estimates = np.zeros(self.n_bootstrap)
        for n_b in tqdm(range(self.n_bootstrap)):            
            # to store the fitted GMM models for each category value
            self.all_gmms = []
            for category_value in range(self.category_values):
                current_ids = np.where(self.y == category_value)
                # we select the relevant latents again
                current_latents = self.X[current_ids]
                current_latents = np.reshape(current_latents, (-1, 1))
                n_components = self.category_best_params[category_value]['bc']
                w_init = self.category_best_params[category_value]['w']
                m_init = self.category_best_params[category_value]['m']
                p_init = self.category_best_params[category_value]['p']
                random_state = self.category_best_params[category_value]['seed']
                if do_bootstrap:
                    # we use index n_b to change the seed so that results will be fully reproducible
                    rng = np.random.default_rng(n_b)
                    X_bs = rng.choice(current_latents, current_latents.shape[0])
                else:
                    X_bs = current_latents 
                gmm = single_fit(X=X_bs, n_components=n_components, reg_covar=self.reg_covar, 
                                 threshold_fit=self.threshold_fit, max_iter=self.max_iter, random_state=random_state, 
                                 w_init=w_init, m_init=m_init, p_init=p_init)               
                self.all_gmms.append(gmm)
            MI_estimates[n_b] = self._calculate_MI_categorical()

        MI_mean = np.mean(MI_estimates)
        MI_std = np.sqrt(np.var(MI_estimates, ddof=1)) if do_bootstrap else None
        return MI_mean, MI_std
       
    def fit_categorical(self, X, y, verbose=False):
        """Calculate mutual information (MI) distribution between a continuous and a categorical variable.
        The first part performs density estimation of the conditional distributions, using GMMs and k-fold cross-validation.
        The second part uses the fitted models to calculate MI, using Monte Carlo integration (numerical integration is not implemented).
        The MI uncertainty is calculated through bootstrap. Loss curves are not returned.
        The complete formula can be found in Appendix B of Piras et al. (2022).

        Parameters
        ----------  
        X : array-like of shape (n_samples)
            Samples of the continuous variable.
        y : array-like of shape (n_samples)   
            Categorical values corresponding to X.
        verbose : bool, default=False
            Whether to print useful procedural statements.
            
        Returns
        ----------
        MI_mean : float
            Mean of the MI distribution, in nat.
        MI_std : float
            Standard deviation of the MI distribution, in nat.
        """
        assert len(X.shape) == 1, f"The shape of the continuous data must be (n_samples), found {X.shape}"    
        assert len(y.shape) == 1, f"The shape of the categorical data must be (n_samples), found {y.shape}"
        assert X.shape[0] == y.shape[0], f"The length of the categorical and continuous data must be the same!"
        self.X = X           
        self.y = y 
        self.category_values = len(np.unique(y))
        self.verbose = verbose

        self.category_best_params = [] # used to store parameters of each GMM fit
        for category_value in range(self.category_values):
            current_ids = np.where(y == category_value)
            # select latents corresponding to current category value
            current_latents = X[current_ids]
            # need to fit the current latents; this is p(z_i | f_i = category_value)
            current_latents = np.reshape(current_latents, (-1, 1))
            
            # initialize attributes every time
            self.converged = False
            self.best_metric = -np.inf
            self.patience_counter = 0
            self.results_dict = {}  
        
            for n_components in range(1, self.max_components+1):
                if self.fixed_components:
                    if n_components < self.fixed_components_number:
                        continue  
                    else:
                        self.converged = True
                current_results_dict = CrossValidation(n_components=n_components, n_folds=self.n_folds, 
                                                       max_iter=self.max_iter, init_type=self.init_type, scale=self.scale,
                                                       n_inits=self.n_inits, threshold_fit=self.threshold_fit, 
                                                       reg_covar=self.reg_covar).fit(current_latents)
                self.results_dict[n_components] = current_results_dict
                if not self.converged:
                    self.metric = self._select_best_metric(n_components=n_components)
                    self._check_convergence(n_components=n_components)

                if self.converged:
                    best_components, best_seed, w_init, m_init, p_init = self._extract_best_parameters(n_components=n_components,                                                                                                       fixed_components=self.fixed_components,
                                                                                            patience=self.patience)
                    # save the best parameters for the current GMM category, and collect all of them before proceeding
                    self.category_best_params.append({'w': w_init, 'm': m_init, 'p': p_init, 
                                                 'bc': best_components, 'seed': best_seed})                    
                    break
            
            # in the unlikely case that we have not reached enough GMM components
            # we raise an error since we need all GMM parameters to calculate MI
            if n_components == self.max_components:
                raise ValueError(f"Convergence in the number of GMM components was not reached! "\
                                 f"Try increasing max_components or threshold_components, "\
                                 f"or decreasing the patience.")

        if self.verbose:
            print(f'Found best parameters for all GMMs, now onto the MI estimation') 
        
        # get MI distribution
        MI_mean, MI_std = self._perform_bootstrap_categorical()
        
        return MI_mean, MI_std 
        



import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.patches as mpatches

def choose_ax(ax, figsize=(15, 10)):
    """Select panel where to plot. Normally returns the input panel; 
    if that is None, create a new panel with the specified figsize. 
    
    Parameters
    ----------
    ax : instance of the axes.Axes class from pyplot
        The input panel.
    figsize : tuple of two integers, default=(15,10)
        The width and height of a newly created panel, in inches.
    
    Returns
    -------
    fig: instance of the figure.Figure class from pyplot
        The output figure.
    ax : instance of the axes.Axes class from pyplot
        The output panel.
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    else:
        fig = None
    return fig, ax


def set_ticksize(ax, axis='both', which='major', labelsize=20, size=6):
    """Set the size of the ticks in a panel.   
    
    Parameters
    ----------
    ax : instance of the axes.Axes class from pyplot
        The panel whose ticks need to be set.
    axis : {'both', 'x', 'y'}, default='both'
        The axis (or axes) to which the parameters are applied.
    which : {'major', 'minor', 'both'}, default='major'
        The group of ticks to which the parameters are applied.
    labelsize : int, default=20
        Font size of the tick label.
    size : int, default=6
        Size of the ticks.
    
    Returns
    -------
    None
    """
    ax.tick_params(axis=axis, which=which, labelsize=labelsize, size=size)

    
def set_titles(ax, xlabel='', ylabel='', title='', fontsize=30, color='black'):
    """Set the titles in a panel.   
    
    Parameters
    ----------
    ax : instance of the axes.Axes class from pyplot
        The panel whose titles need to be set.
    xlabel : string, default=''
        The x-label.
    ylabel : string, default=''
        The y-label.
    title : string, default=''
        The title of the panel.
    fontsize : int, default=30
        Font size of all titles and labels.
    color : string, default='black'
        Color of all titles and labels.
    
    Returns
    -------
    None
    """
    ax.set_xlabel(xlabel, fontsize=fontsize, color=color)
    ax.set_ylabel(ylabel, fontsize=fontsize, color=color)
    ax.set_title(title, fontsize=fontsize, color=color)
    

def set_legend(ax, fontsize=25, frameon=False, loc='best', handles=None, labels=None):
    """Set the legend in a panel.   
    
    Parameters
    ----------
    ax : instance of the axes.Axes class from pyplot
        The panel whose legend need to be set.
    fontsize : int, default=25
        Font size of the legend.
    frameon : bool, default=False
        Whether to frame the legend or not.
    loc : string, default='best'
        Position of the legend.   
    handles : sequence of `.Artist` (see pyplot docs), default=None
        The handles, only used when playing around with the order of the legend (quite rarely),
        and together with labels.
    labels : list of str, default=None
        The labels, only used when playing around with the order of the legend (quite rarely), 
        and together with handles.
    
    Returns
    -------
    None
    """
    if handles is not None:
        ax.legend(handles, labels, fontsize=fontsize, frameon=frameon, loc=loc)            
    else:
        ax.legend(fontsize=fontsize, frameon=frameon, loc=loc)    

    
def calculate_ellipse_params(covariance):
    """Given a covariance matrix in 2D, calculate the orientation (angle), height and width of 
    the error ellipse representing an iso-contour of the corresponding Gaussian distribution.
    Based on https://www.visiondummy.com/2014/04/draw-error-ellipse-representing-covariance-matrix/.
    
    Parameters
    ----------
    covariance : array-like of shape (2, 2)
        The covariance matrix of the Gaussian distribution.

    Returns
    -------
    angle : float
        Orientation of the error ellipse.
    width : float
        Width of the error ellipse.
    height : float
        Height of the error ellipse.
    """
    U, s, Vt = np.linalg.svd(covariance)
    # using arctan2 instead of arctan needed to take into account sign
    angle = np.degrees(np.arctan2(U[1, 0], U[0, 0]))
    width, height = 2*np.sqrt(s)
    return angle, width, height
    
    
def find_contours(contour_levels):
    """Associate the user-required sigma levels to the actual size of the error ellipse.
    The 1-sigma, 2-sigma, and/or 3-sigma contours are obtained from a chi2 table, and based on
    https://www.visiondummy.com/2014/04/draw-error-ellipse-representing-covariance-matrix/.
    
    Parameters
    ----------
    contour_levels : list
        The user-required contours to display, in sigmas.
        Only 1, 2 and 3 are accepted in this list.

    Returns
    -------
    ranges : list
        The size of the error ellipse corresponding to the input contour levels.
    """
    if any(cl not in [1,2,3] for cl in contour_levels):
        raise ValueError(f'Contour levels must be either 1, 2 and 3; current contour levels are {contour_levels}.')
    assert len(contour_levels) <= 3, 'Length of contour levels cannot exceed 3'
    ranges_dict = {1: np.sqrt(0.77), 2: np.sqrt(5.991), 3: np.sqrt(10.597)}
    ranges = []
    for cl in contour_levels:
        ranges.append(ranges_dict[cl])
    return ranges
        
    
def draw_ellipse(mean, covariance, weight, ax=None, contour_levels=[2],
                 weight_size=1000, alpha=None, color=None, label=None, marker='X',
                 component_count=0, legendsize=30, loc='best', frameon=False, **kwargs):
    """Draw the error ellipse corresponding to a given mean vector and Gaussian covariance matrix.
    Since this is don in the context of Gaussian mixture models (GMMs), the weight is also indicated 
    with a marker whose size is proportional to the weight value. Based on
    https://www.visiondummy.com/2014/04/draw-error-ellipse-representing-covariance-matrix/.
    
    Parameters
    ----------
    mean : array-like of shape (2)
        The mean of the Gaussian distribution.
    covariance : array-like of shape (2, 2)
        The covariance matrix of the Gaussian distribution.
    weight : float
        The weight associated with this GMM component.
    ax : instance of the axes.Axes class from pyplot, default='None'
        The panel where to draw the ellipse.    
    contour_levels : list of integers, default=[2]
        The sigma levels whose contour are going to be drawn. 
        The integers can be only 1, 2 and 3.
    weight_size : float, default=1000
        The multiplier for the marker size (also proportional to the weight).   
    alpha : float, default=None
        Transparency, between 0 and 1.    
    color : 'string', default=None
        Color of the ellipse.
    marker : string, default='X'
        Marker type for the centroid of the ellipse.
    label : string, default=None
        Label associated to the ellipse.
    component_count : int, default=0
        A counter so that the label is printed only for the first component.
    legendsize : int, default=30
        Font size of the legend.
    loc : string, default='best'
        Legend position.
    frameon : bool, default=False
        Whether to frame the legend or not.
    kwargs : dictionary
        The extra keyword arguments to pass to the plotting function.    
        
    Returns
    -------
    None
    """
    fig, ax = choose_ax(ax)
    assert covariance.shape == (2, 2), 'Each covariance matrix must be of shape (2, 2)!'
    # convert covariance to the parameters of the corresponding error ellipse
    angle, width, height = calculate_ellipse_params(covariance)
    # find the contour sigma levels
    contours = find_contours(contour_levels)   
    # draw the ellipse
    for contour in contours:
        ax.add_patch(Ellipse(mean, contour*width, contour*height, angle, color=color, fill=False, 
                             alpha=alpha, label=label if component_count==0 else "", **kwargs))
        ax.scatter(mean[0], mean[1],  marker=marker, color=color,
                   s=weight_size*weight, alpha=alpha)
    set_legend(ax, fontsize=legendsize, frameon=frameon)

    
def plot_gmm_contours(gmm, ax=None, color='salmon', alpha=0.8, 
                      linewidth=4, fill=False, label='', xlabel='X1', ylabel='X2',
                      **kwargs):
    """Draw the contour ellipses corresponding to a given Gaussian mixture model (GMM).
    
    Parameters
    ----------
    gmm : instance of the GMM class
        The GMM whose contours are going to be displayed.
    ax : instance of the axes.Axes class from pyplot
        The panel where to plot the samples.   
    color : string, default='salmon'
        The color of the GMM contours.
    alpha : float, default=0.8
        Transparency of the contour lines; must be between 0 and 1.
    linewidth : int, default=4
        The width of the contour lines.
    fill : bool, default=False
        Whether to fill the contours or not.
    label : string, default=''
        The legend label to associate to the contours.
    xlabel : string, default='X1'
        Label of the x-axis.
    ylabel : string, default='X2'
        Label of the y-axis. 
    kwargs : dictionary
        The extra keyword arguments to pass to the plotting functions.
        
    Returns
    -------
    fig: instance of the figure.Figure class from pyplot
        The output figure.
    ax : instance of the axes.Axes class from pyplot
        The output panel.
    """
    fig, ax = choose_ax(ax, figsize=(10, 10))
            
    means, covariances, weights = gmm.means_, gmm.covariances_, gmm.weights_
    for component_count, (mean, covariance, weight) in enumerate(zip(means, covariances, weights)):
        draw_ellipse(mean, covariance, weight, ax=ax, component_count=component_count, alpha=alpha, 
                     color=color, linewidth=linewidth, label=label, **kwargs)  
    set_ticksize(ax)
    set_titles(ax, xlabel=xlabel, ylabel=ylabel)
    return fig, ax


def calculate_summary_and_significance(estimates):
    """Given an array of estimates, calculate their sample mean, the error on the mean,
    and the significance with respect to 0.
    
    Parameters
    ----------
    estimates : 1D array-like
        The value of all estimates for a certain number of trials. 
        
    Returns
    -------
    mean_value : float
        The mean of the estimates.
    mean_error : float
        The error on the mean of the estimates.
    significance : float
        The distance of the mean with respect to 0.
    """
    trials_number = len(estimates)
    mean_value = np.mean(estimates)
    mean_error = np.sqrt(np.var(estimates, ddof=1))/np.sqrt(trials_number)
    # significance is with respect to 0
    significance = np.mean(estimates) / mean_error
    return mean_value, mean_error, significance


def histogram_estimates(estimates, ax=None, bins=20, alpha=1, title='',
                   legendsize=25, color='salmon', lw=3, histtype='step'):
    """Plot the histogram of an array of estimates, centred on 0.
    
    Parameters
    ----------
    estimates : 1D array-like
        The estimates whose histogram we plot. 
    ax : instance of the axes.Axes class from pyplot, default=None
        The panel where to draw the histogram. 
    bins : int, default=20
        Number of histogram bins.
    alpha : float, default=1
        Transparency of the histogram line. Must be between 0 and 1.
    title : string, default=''
        Title of the panel.
    legendsize : int, default=25
         Font size of the legend.
    color : string, default='salmon'
        Histogram color.
    lw : int, default=3
        Line width of the histogram line.  
    histtype : {'bar', 'barstacked', 'step', 'stepfilled'}, default='step'
        The type of histogram to draw.
        
    Returns
    -------
    None
    """    
    fig, ax = choose_ax(ax)
    mean_value, mean_error, significance = calculate_summary_and_significance(estimates)
    ax.hist(estimates.flatten(), alpha=alpha, bins=bins, color=color, histtype=histtype, lw=lw,
            label=f'MI values, {significance:.1f}$\sigma$,\n{mean_value:.4f}$\pm${mean_error:.4f}' 
            )
    ax.axvline(0, ls='--', lw=lw, c='black')
    set_ticksize(ax)
    set_titles(ax, xlabel='MI [nat]', ylabel='Counts', title=title)
    set_legend(ax, fontsize=legendsize)


def swap_leg_handles(ax, i1=-1, i2=-2):
    """Swap legend handles and labels.
    
    Parameters
    ----------
    ax : instance of the axes.Axes class from pyplot
        The panel whose legend is going to be affected
    i1 : int, default=-1
        The first index to swap.
    i2 : int, default=-2
        The first index to swap.
                
    Returns
    -------
    handles : sequence of `.Artist` (see pyplot docs)
        The swapped handles.
    labels : list of str
        The swapped labels.
    """      
    handles, labels = ax.get_legend_handles_labels()
    handles[i1], handles[i2] = handles[i2], handles[i1] 
    labels[i1], labels[i2] = labels[i2], labels[i1] 
    return handles, labels

def plot_multivariate_normal(ax=None, mean=0, cov=1, alpha=1, color='black', lw=5, ls='--'):
    """Plot Gaussian distribution in 1D.
    
    Parameters
    ----------
    ax : instance of the axes.Axes class from pyplot, default=None
        The panel where to draw the Gaussian. 
    mean : float, default=0
        The mean of the Gaussian distribution.
    cov : float, default=1
        The covariance of the Gaussian distribution.        
    alpha : float, default=2
        Transparency of the line. Must be between 0 and 1.
    color : string, default='black'
        Line color.
    lw : int, default=5
        Line width. 
    ls : string, default='--'
        Line style. Check pyplot for possible values.
        
    Returns
    -------
    None
    """
    from scipy.stats import multivariate_normal
    fig, ax = choose_ax(ax)
    x = np.linspace(mean-3, mean+3, 1000, endpoint=False)
    y = multivariate_normal.pdf(x, mean=mean, cov=cov)
    ax.plot(x, y, c=color, ls=ls, lw=lw, alpha=alpha, label=r'$\mathcal{N}(0, 1)$')

    
def plot_bias_chi2_histogram(estimates, analytic, ax=None, bins=20, alpha=0.5, show_title=True, show_legend=True,
                             legendsize=44, color='grey', lw=3, histtype='bar', savefig=False, savepath='./', 
                             labelfontsize=60, labelcolor='black', hide_ylabel=False, ds=''):
    """Plot the histogram of the bias of estimates, centred on 0.
    Also calculate reduced chi2.
    
    Parameters
    ----------
    estimates : array-like of shape (n_estimates, 2)
        The estimates whose bias we histogram. 
        First column is the MI estimate,
        second column is the MI error.
    analytic : float
        The analytic MI value.
    ax : instance of the axes.Axes class from pyplot, default=None
        The panel where to draw the histogram. 
    bins : int, default=20
        Number of histogram bins.
    alpha : float, default=0.5
        Transparency of the histogram line. Must be between 0 and 1.
    show_title : bool, default=True
        Whether to display the title with bias and chi2 results.
    show_legend : bool, default=True
        Whether to display the legend.
    legendsize : int, default=44
         Font size of the legend.
    color : string, default='grey'
        Histogram color.
    lw : int, default=3
        Line width of the histogram line.  
    histtype : {'bar', 'barstacked', 'step', 'stepfilled'}, default='bar'
        The type of histogram to draw.
    savefig : bool, default=False
        Whether to save or not the figure as a pdf file (for the paper).
    savepath : string, default='./'
        The path where to save the figure. Only used if savefig is True.  
    labelfontsize : int, default=60
        Fontsize of labels and title.
    labelcolor : string, default='black'
        Color of labels and title.
    hide_ylabel : bool, default=False
        Whether to paint the ylabel white, so that it's hidden.
        Only for the paper.
    ds : string, default=''
        The dataset which is being analysed.
        
    Returns
    -------
    biases : np.array of shape (n_estimates)
        The biases of the estimates.
    red_chi2 : float
        The reduced chi2 of the estimates.
    df : int
        The number of degrees of freedom.
    """  
    biases = np.array((estimates[:, 0]-analytic)/(estimates[:, 1]))
    chi2 = np.sum(biases**2)
    df = len(estimates)-1    
    red_chi2 = chi2/df
    fig, ax = choose_ax(ax)
    ax.hist(biases, alpha=alpha, bins=bins, 
             label='GMM-MI', color=color, density='True')
    plot_multivariate_normal(ax=ax)    
    if show_title:
        title = f'{ds}: {np.mean(biases):.2f}$\pm${np.std(biases):.2f}, $\chi^2_{{{{red}}}}$={red_chi2:.2f} with {df} dof'
    else:
        title = ''
    set_titles(ax, xlabel='Standardized residual', ylabel='Probability', title=title, 
               fontsize=labelfontsize, color=labelcolor)
    if hide_ylabel:
        ax.set_ylabel(ylabel='Probability', fontsize=labelfontsize, color='white')
    if show_legend:
        handles, labels = swap_leg_handles(ax, i1=-1, i2=-2)
        set_legend(ax, fontsize=legendsize, handles=handles, labels=labels, frameon=False, loc='upper left')
    set_ticksize(ax, labelsize=45, size=12)
    if savefig:
        plt.savefig(savepath, bbox_inches='tight')
    return biases, red_chi2, df
        
    
def plot_loss_curve(loss_curves, index, ax):
    """Plot individual panel (referenced by an index) of the loss curves.
    
    Parameters
    ----------
    loss_curves : list of lists
        All loss curves, as returned by GMM-MI. Every pair of elements represents the train 
        and validation loss, respectively, for a particular fold and initialization.
    index : int
        The panel index where to draw the loss curves.
    ax : instance of the axes.Axes class from pyplot
        The current panel where to draw the loss curves. 
        
    Returns
    -------
    None
    """    
    current_train_loss = loss_curves[2*index]
    current_val_loss = loss_curves[2*index+1]
    iterations = np.arange(1, 1+len(current_train_loss))

    ax.plot(iterations, current_train_loss, label='Train', lw=3, ls='-', color='grey')
    ax.plot(iterations, current_val_loss, label='Valid', lw=3, ls='-', color='salmon')
    
def plot_loss_curves(loss_curves, n_inits, n_folds, figsize=(20, 30), 
                     legendsize=18, frameon=False, loc='lower right'):
    """Plot all loss curves for the k-fold cross-validation in GMM-MI. The legend is placed only
    in the first panel. Every row corresponds to a different initialization, and every column
    corresponds to a different fold.
    
    Parameters
    ----------
    loss_curves : list of lists
        All loss curves, as returned by GMM-MI. Every pair of elements represents the train 
        and validation loss, respectively, for a particular fold and initialization.
    n_inits : int
        The number of initializations in the k-fold cross-validation.
    n_folds : int
        The number of folds in the k-fold cross-validation (k = n_folds).
    figsize : tuple of two integers, default=(20,30)
        The width and height of the entire plot, in inches.
    legendsize : int, default=18
        Font size of the legend (drawn only in the first panel).
    frameon : bool, default=False
        Whether to frame the legend or not.    
    loc : string, default='lower right'
        Position of the legend.     
        
    Returns
    -------
    None
    """
    fig, axes = plt.subplots(n_inits, n_folds, figsize=figsize, sharex=True, sharey=True)
    axes = axes.flatten()
    
    for i in range(n_inits*n_folds):
        ax = axes[i]
        plot_loss_curve(loss_curves, i, ax)
        if i == 0:
            set_legend(ax, fontsize=legendsize, frameon=frameon, loc=loc)
        if i >= (n_inits-1)*n_folds and i % n_folds != 0:
            set_titles(ax, xlabel='Iteration', fontsize=20)
        elif i % n_folds == 0 and i < (n_inits-1)*n_folds:
            set_titles(ax, ylabel='logL', fontsize=20)
        elif i % n_folds == 0 and i >= (n_inits-1)*n_folds:
            set_titles(ax, xlabel='Iteration', ylabel='logL', fontsize=20)            
        set_ticksize(ax)        
    fig.subplots_adjust(wspace=0.0, hspace=0.0)


def create_heatmap(data, row_labels, col_labels, fsize=30, ax=None,
            cbar_kw={}, cbarlabel="", **kwargs):
    """
    Create a heatmap from a numpy array and two lists of labels.

    Parameters
    ----------
    data : array-like of shape (n_rows, n_columns)
        A 2D numpy array containing the data values that are reported.
    row_labels : list
        A list with the labels for the rows.
    col_labels : list
        A list with the labels for the columns.
    ax : `matplotlib.axes.Axes` instance, default=None
        The panel in which the heatmap is plotted. 
        If not provided, use current axes or create a new one.
    cbar_kw : dict, default={}
        Contains the arguments to `matplotlib.Figure.colorbar`.
    cbarlabel : string, default=""
        The label for the colorbar.
    **kwargs
        All other arguments are forwarded to `imshow`.
        
    Returns
    -------
    im : AxesImage object
        The output of `imshow`.
    cbar : `colorbar` object
        The colorbar.
    """

    fig, ax = choose_ax(ax, figsize=(24, 8))
       
    # plot the heatmap
    im = ax.imshow(data, **kwargs)
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.5)
    # create colorbar
    cbar = ax.figure.colorbar(im, ax=ax, cax=cax, **cbar_kw)
    cbar.ax.tick_params(labelsize=20, length=10, width=2) 
    cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom", fontsize=30)

    # show all ticks and label them with the respective list entries.
    ax.set_xticks(np.arange(data.shape[1]))
    ax.set_xticklabels(labels=col_labels, fontsize=fsize)
    ax.set_yticks(np.arange(data.shape[0]))
    ax.set_yticklabels(labels=row_labels, fontsize=fsize)

    # set the horizontal axes labeling appear on top.
    ax.tick_params(top=True, bottom=False,
                   labeltop=True, labelbottom=False, length=10, width=2)

    # Turn spines off and create white grid.
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['bottom'].set_visible(False)  
    ax.set_xticks(np.arange(data.shape[1]+1)-.5, minor=True)
    ax.set_yticks(np.arange(data.shape[0]+1)-.5, minor=True)
    ax.tick_params(which="minor", bottom=False, left=False)

    return im, cbar


def annotate_heatmap(im, data=None, valfmt="{x:.2f}",
                     textcolors=("black", "white"),
                     threshold=None, save_fig=False,
                     save_path="", **textkw):
    """Used to annotate a previously-created heatmap.

    Parameters
    ----------
    im : AxesImage
        The image to be annotated.
    data : array-like of shape (n_rows, n_columns)
        Data used to annotate.  If None, the image's data is used.
    valfmt : either str, or `matplotlib.ticker.Formatter`, default="{x:.2f}"
        The format of the annotations inside the heatmap.  This should either
        use the string format method, e.g. "$ {x:.2f}", or be a
        `matplotlib.ticker.Formatter`.
    textcolors : tuple, default=("black", "white")
        A pair of colors.  The first is used for values below a threshold,
        the second for those above.
    threshold : float, default=None
        Value in data units according to which the colors from textcolors are
        applied.  If None (the default) uses the middle of the colormap as
        separation.
    save_fig : bool, default=False
        Whether to save the annotated heatmap or not.
    save_path : string, default=""
        If save_fig is True, the path where to save the figure.
    **kwargs
        All other arguments are forwarded to each call to `text` used to create the text labels.
        
    Returns
    -------
    texts : list
        Contains all the text that is annotated onto the heatmap.
    """

    if not isinstance(data, (list, np.ndarray)):
        data = im.get_array()
        
    # normalize the threshold to the images color range.
    if threshold is not None:
        threshold = im.norm(threshold)
    else:
        threshold = im.norm(data.max())/2.

    # set default alignment to center, but allow it to be overwritten by textkw.
    kw = dict(horizontalalignment="center",
              verticalalignment="center")
    kw.update(textkw)

    # get the formatter in case a string is supplied
    if isinstance(valfmt, str):
        valfmt = matplotlib.ticker.StrMethodFormatter(valfmt)

    # loop over the data and create a `Text` for each "pixel".
    # change the text's color depending on the data.
    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            # ignore if value is too small
            if data[i, j] < 0.01:
                text = im.axes.text(j, i, matplotlib.ticker.StrMethodFormatter("{x:.0f}")(data[i, j], None), **kw) 
            else:
                text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
            texts.append(text)
            
    if save_fig:
        plt.savefig(save_path, bbox_inches='tight')
    return texts


def add_label_violin_plot(violin, label, labels):
    """ Add labels to violin plot legend.
    
    Parameters
    ----------
    violin : instance of axes.Axes.violinplot class
        The violin plot to add to the legend.
    label : str
        The label for the current violin plot.
    labels : list
        List containing all labels for all violins
        
    Returns
    -------
    labels : list
        The updated list of labels.
    """
    color = violin["bodies"][0].get_facecolor().flatten()
    labels.append((mpatches.Patch(color=color), label))
    return labels

def violin_plots(y_values, std_values, transformation, N_values,
                 ksg_values=None, alpha=1, color='#d95f02', fontsize=27, widths=0.4):
    """
    
    Parameters
    ----------
    y_values : array of shape (len(N_values), random_states)
        Value of MI as returned by GMM-MI.
    std_values : array of shape (len(N_values), random_states)
        Values of MI standard deviations, as returned by GMM-MI.
    transformation : str
        The transformation that the data was subject to.
    N_values : list
        Contains the number of samples.
    ksg_values : array of shape (len(N_values), random_states), default=None
        Contains the KSG values, for comparison.    
    alpha : float, default=1
        Transparency, between 0 and 1.    
    color : 'string', default='#d95f02'
        Color of the GMM results.
    fontsize : int, default=27
        Size of label font.
    widths : float, default=0.4
        The percentage width of each violin plot; must be between 0 and 1.
        
    Returns
    -------
    None
    """
    labels = [] # for the legend
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(8, 10), sharex=True)
    # horizontal line
    ax1.axhline(0, linewidth=3, c='black', linestyle='--')
    ksg = False if ksg_values is None else True
    
    # first violin
    if ksg:
        parts = ax1.violinplot(y_values.T, positions=np.arange(1, 5, 1.5), showmeans=True, widths=widths)
    else:
        parts = ax1.violinplot(y_values.T, showmeans=True, widths=widths)
        ax1.set_ylim((-0.19, 0.29))
    for pc in parts['bodies']:
        pc.set_facecolor(color)
        pc.set_edgecolor('black')
        pc.set_alpha(0.8)
    for partname in ('cbars','cmins','cmaxes','cmeans'):
        vp = parts[partname]
        vp.set_edgecolor('black')
        vp.set_linewidth(1.2)    
    labels = add_label_violin_plot(parts, 'GMM-MI', labels)  
    xy = [[l.vertices[:,0].mean(),l.vertices[0,1]] for l in parts['cmeans'].get_paths()]
    xy = np.array(xy)
    ax1.scatter(xy[:,0], xy[:,1], s=121, c="black", marker="x", zorder=3)
    # make lines invisible
    parts['cmeans'].set_visible(False) 

    # possible ksg violin
    if ksg:
        parts = ax1.violinplot(ksg_values.T, positions=np.arange(0.5, 5, 1.5), showmeans=True, widths=widths)
        for pc in parts['bodies']:
            pc.set_facecolor('grey')
            pc.set_edgecolor('black')
            pc.set_alpha(0.8)
        for partname in ('cbars','cmins','cmaxes','cmeans'):
            vp = parts[partname]
            vp.set_edgecolor('black')
            vp.set_linewidth(1.2)    
        labels = add_label_violin_plot(parts, 'KSG', labels)  
        xy = [[l.vertices[:,0].mean(),l.vertices[0,1]] for l in parts['cmeans'].get_paths()]
        xy = np.array(xy)
        ax1.scatter(xy[:,0], xy[:,1], s=121, c="black", marker="x", zorder=3)
        # make lines invisible
        parts['cmeans'].set_visible(False) 
    
    # second violin
    if ksg:
        parts = ax2.violinplot(std_values.T, positions=np.arange(1, 5, 1.5), showmeans=True, widths=widths)
    else:
        parts = ax2.violinplot(std_values.T, showmeans=True, widths=widths)
    for pc in parts['bodies']:
        pc.set_facecolor(color)
        pc.set_edgecolor('black')
        pc.set_alpha(0.8)
    for partname in ('cbars','cmins','cmaxes','cmeans'):
        vp = parts[partname]
        vp.set_edgecolor('black')
        vp.set_linewidth(1.2)   
    xy = [[l.vertices[:,0].mean(),l.vertices[0,1]] for l in parts['cmeans'].get_paths()]
    xy = np.array(xy)
    ax2.scatter(xy[:,0], xy[:,1], s=121, c="black", marker="x", zorder=3)
    # make lines invisible
    parts['cmeans'].set_visible(False) 
       
    # draw 1/sqrt(N) line
    pivot = np.sqrt(N_values[2])*np.mean(std_values[2])
    std_trend_values = [pivot/np.sqrt(N) for N in N_values]
    if ksg:
        ax2.plot(np.arange(1, 5, 1.5), std_trend_values, color='grey', 
             label='$\\propto \\frac{1}{\sqrt{N}}$', lw=2, alpha=0.5)        
    else:
        ax2.plot(np.arange(1, 4), std_trend_values, color='grey', 
             label='$\\propto \\frac{1}{\sqrt{N}}$', lw=2, alpha=0.5)
    ax2.set_yscale('log')
    ax2.set_ylim((10**(-2.5), 10**(-0.75))) 
    
    # final touches
    set_ticksize(ax1)
    set_ticksize(ax2)
    if ksg:
        ax2.set_xticks([0.75, 2.25, 3.75]) 
    else:
        ax2.set_xticks(np.arange(1, 4)) 
    ax2.set_xticklabels(N_values, fontsize=fontsize)
    ax2.set_xlabel('$N$', fontsize=fontsize)
    ax1.set_ylabel('Bias [nat]', fontsize=fontsize)
    ax2.set_ylabel('$\log_{10}(\sigma)$', fontsize=fontsize)
    ax1.legend(*zip(*labels), loc='upper right', frameon=False, fontsize=25)
    ax2.legend(loc='upper right', frameon=False, fontsize=28)

    fig.align_ylabels()
    fig.subplots_adjust(wspace=0, hspace=0)
    if ksg:
        plt.suptitle(f'{transformation.capitalize()}', fontsize=25)
    else:
        fig.savefig(f'./figures/{transformation}_transformation.pdf', bbox_inches='tight')


from setuptools import setup, find_packages
import shutil

PACKAGENAME = 'gmm_mi'

setup(
    name='gmm_mi',
    version="0.1.0",
    author='Davide Piras',
    author_email='dr.davide.piras@gmail.com',
    description='Estimate mutual information distribution with Gaussian mixture models',
    url='https://github.com/dpiras/GMM-MI',
    license='GNU General Public License v3.0 (GPLv3)',
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    install_requires=['pytest',
                      'tqdm',
                      'matplotlib>=3.1.2',
                      'numpy>=1.17.4',
                      'scikit-learn>=1.0.2',
                      'scipy>=1.7.1']
                      )

